/*
   makerx.c - create an RX01 disk image with bootable fig-FORTH.
   Includes an RT-11 directory structure with FORTH.DAT and FORTH.MAC.

   Author: Paul Hardy

   This code is in the public domain.

   Synopsis: makerx > raw_rx01_image.dsk

   The RT-11 directory starts at logical byte number 6000 octal.
   The first file in the RT-11 file system starts at logical byte
   number 16000.
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Macros for high and low byte in a word */
#define LOW(X)  ( X       & 0xFF)
#define HIGH(X) ((X >> 8) & 0xFF)

/*
   Disk size: an RX01 has 77 tracks, 26 sectors/track, 128 bytes/sector,
   which is 256,256 bytes.  Track 0 is reserved; we start writing at
   track 1, so we only use 76 tracks (1..76).
*/
#define DISK_BYTES (76 * 26 * 128)


int main () {
   uint8_t  rx01[DISK_BYTES]; /* RX01 disk image */
   int      i,j;            /* loop variables */
   int      addr;           /* byte location on disk */
   int      inchar;         /* input character */
   uint8_t  outch1, outch2; /* 2 bytes for little-endian output */
   uint16_t outword;        /* 2 byte PDP-11 word */
   uint16_t fname1, fname2, fext;  /* file name and extension in RADIX 50    */
   uint16_t sum16;                 /* Files-11 checksum                      */
   int      datblocks, macblocks;  /* disk blocks in FORTH.DAT and FORTH.MAC */
   int      freeblocks;            /* number of free blocks left on disk     */
   int      datbytes, macbytes;    /* bytes in FORTH.DAT and FORTH.MAC       */
   uint16_t cdate = (1 << 10) | (21 << 5) | (1980 - 1972); /* 21-Jan-1980    */
   FILE     *infp;  /* input file pointer */

   int sequence, track, sector; /* for skewing calculations */

   uint8_t  raw01[DISK_BYTES]; /* RX01 copy, for skewing */

   void ascii_rad50 (char *, uint16_t *);

   FILE *fopen (const char *, const char *);
   int  fclose (FILE *);
   int  fseek(FILE *stream, long offset, int whence);
   long ftell(FILE *stream);
   void rewind(FILE *stream);

   for (i = 0; i < DISK_BYTES; i++) rx01[i] = 0;

   /*
      Write Home Block (Block 1)
   */
   rx01 [01722] = 1;  /* Cluster size */
   rx01 [01724] = 6;  /* Block number of first directory segment */
   ascii_rad50 ("V3A", &fname1);  /* System Version */
   rx01 [01726] = LOW (fname1);  rx01 [01727] = HIGH (fname1);
   /* Volume Identification */
   strncpy ((char *)&rx01 [01730], "FORTH-1.3.1 ", 12);
   /* Owner Name            */
   strncpy ((char *)&rx01 [01744], "F.I.G.      ", 12);
   /* System Identification */
   strncpy ((char *)&rx01 [01760], "DECRT11A    ", 12);
   /*
      Compute checksum for home block.
   */
   sum16 = 0;
   for (i = 01000; i < 01776; i += 2) {
      sum16 += rx01 [i] | (rx01 [i+1] << 8);
   }
   rx01 [01776] = LOW (sum16);  rx01 [01777] = HIGH (sum16);

   freeblocks = 0740;  /* space on disk from byte 16000 with no files */

   /*
      Process FORTH.DAT
   */
   addr = 016000; /* start of first file, FORTH.DAT */

   if ((infp = fopen ("FORTH.DAT", "r")) == NULL) {
      fprintf (stderr, "\nERROR: cannot open FORTH.DAT\n\n");
      exit (EXIT_FAILURE);
   }
   fseek (infp, 0L, SEEK_END);
   datbytes = ftell (infp);
   datblocks = (datbytes + 0777) / 01000; /* for directory entry */
   rewind (infp);
   for (i = 0; i < datbytes; i++) {
      inchar = fgetc (infp) & 0xFF;
      rx01 [addr++] = inchar;
   }
   (void)fclose (infp);
   /* Pad last block of FORTH.DAT with spaces, if necessary */
   /* FORTH.DAT should always end on a byte block */
   while ((i & 0777) != 0) {
      rx01 [addr++] = ' ';
      i++;
   }

   freeblocks -= datblocks;

   /*
      Process FORTH.MAC
   */
   addr = (addr + 0777) & 0777000; /* jump to start of next block */
   if ((infp = fopen ("FORTH.MAC", "r")) == NULL) {
      fprintf (stderr, "\nERROR: cannot open FORTH.MAC\n\n");
      exit (EXIT_FAILURE);
   }
   fseek (infp, 0L, SEEK_END);
   macbytes = ftell (infp);
   macblocks = (macbytes + 0777) / 01000; /* for directory entry */
   rewind (infp);
   for (i = 0; i < macbytes; i++) {
      inchar = fgetc (infp) & 0xFF;
      rx01 [addr++] = inchar;
   }
   (void)fclose (infp);
   /* Pad last block of FORTH.MAC with nulls for RT-11, if necessary */
   while ((i & 0777) != 0) {
      rx01 [addr++] = '\0';
      i++;
   }

   freeblocks -= macblocks;

   /*
      Create directory segments; write in little-endian order.
   */
   addr = 06000;  /* Location is 006000 */
   rx01[addr++] = 004; rx01[addr++] = 000; /* 4 directory segments           */
   rx01[addr++] = 000; rx01[addr++] = 000; /* no next directory segment      */
   rx01[addr++] = 001; rx01[addr++] = 000; /* highest directory segment used */
   rx01[addr++] = 000; rx01[addr++] = 000; /* no information bits in entries */
   rx01[addr++] = 016; rx01[addr++] = 000; /* first file starting block      */

   /*
      Write FORTH.DAT directory entry.
   */
   /* Location is 006012 */
   /* Mark entry as a permanent file */
   rx01[addr++] = LOW (02000); rx01[addr++] = HIGH (02000);
   /* Write the file name and extension in RADIX 50 encoding */
   ascii_rad50 ("FOR", &fname1);
   ascii_rad50 ("TH ", &fname2);
   ascii_rad50 ("DAT", &fext);
   rx01[addr++] = LOW (fname1); rx01[addr++] = HIGH (fname1);
   rx01[addr++] = LOW (fname2); rx01[addr++] = HIGH (fname2);
   rx01[addr++] = LOW (fext);   rx01[addr++] = HIGH (fext);
   /* Write the file length, in blocks */
   rx01[addr++] = LOW (datblocks); rx01[addr++] = HIGH (datblocks);
   /* File is not open; write two 0 bytes as status */
   rx01[addr++] = 000; rx01[addr++] = 000;
   /* Write creation date as 21-Jan-1980 */
   rx01[addr++] = LOW (cdate); rx01[addr++] = HIGH (cdate);

   /*
      Write FORTH.MAC directory entry.
   */
   /* Location is 006026 */
   /* Mark entry as a permanent file */
   rx01[addr++] = LOW (02000); rx01[addr++] = HIGH (02000);
   /* Write the file name and extension in RADIX 50 encoding */
   ascii_rad50 ("FOR", &fname1);
   ascii_rad50 ("TH ", &fname2);
   ascii_rad50 ("MAC", &fext);
   rx01[addr++] = LOW (fname1); rx01[addr++] = HIGH (fname1);
   rx01[addr++] = LOW (fname2); rx01[addr++] = HIGH (fname2);
   rx01[addr++] = LOW (fext);   rx01[addr++] = HIGH (fext);
   /* Write the file length, in blocks */
   rx01[addr++] = LOW (macblocks); rx01[addr++] = HIGH (macblocks);
   /* File is not open; write two 0 bytes as status */
   rx01[addr++] = 000; rx01[addr++] = 000;
   /* Write creation date as 21-Jan-1980 */
   rx01[addr++] = LOW (cdate); rx01[addr++] = HIGH (cdate);

   /*
      Write EMPTY.FIL directory entry.
   */
   /* Location is 006042 */
   /* Mark entry as an empty file */
   rx01[addr++] = LOW (01000); rx01[addr++] = HIGH (01000);
   /* Write the file name and extension in RADIX 50 encoding */
   ascii_rad50 (" EM", &fname1);
   ascii_rad50 ("PTY", &fname2);
   ascii_rad50 ("FIL", &fext);
   rx01[addr++] = LOW (fname1); rx01[addr++] = HIGH (fname1);
   rx01[addr++] = LOW (fname2); rx01[addr++] = HIGH (fname2);
   rx01[addr++] = LOW (fext);   rx01[addr++] = HIGH (fext);
   /* Write the free space left on disk, in blocks */
   rx01[addr++] = LOW (freeblocks); rx01[addr++] = HIGH (freeblocks);
   /* File is not open; write two 0 bytes as status */
   rx01[addr++] = 000; rx01[addr++] = 000;
   /* Write blank creation date */
   rx01[addr++] = 000; rx01[addr++] = 000;

   /*
      Mark the end of this directory segment.
   */
   rx01[addr++] = LOW (04000); rx01[addr++] = HIGH (04000);

   /*
      Write binary data: boot sector and Forth binary in FORTH.DAT.
   */
   /* Boot sector: Track 1, Sector 1 and start of FORTH.DAT Screen 40 */
   if ((infp = fopen ("boot1.bin", "r")) == NULL) {
      fprintf (stderr, "ERROR: Cannot open boot1.bin for input.\n\n");
      exit (EXIT_FAILURE);
   }
   for (i = 0; i < 128 && !feof (infp); i++) {
      inchar = fgetc (infp);
      if (!feof (infp)) {
         rx01 [i] =                          /* Track 1, Sector 1   */
         rx01 [0134000 + i] = inchar & 0xFF; /* FORTH.DAT Screen 40 */
      }
   }
   fclose (infp);

   /* Forth binary */
   if ((infp = fopen ("forths.bin", "r")) == NULL) {
      fprintf (stderr, "ERROR: Cannot open forths.bin for input.\n\n");
      exit (EXIT_FAILURE);
   }
   for (i = 01000; i < (8 * 1024 - 256) && !feof (infp); i++) {
      inchar = fgetc (infp);
      if (!feof (infp)) rx01 [0134000 + i] = inchar & 0xFF;
   }
   fclose (infp);

   /* 2-sector boot loader */
   if ((infp = fopen ("boot2.bin", "r")) == NULL) {
      fprintf (stderr, "ERROR: Cannot open boot2.bin for input.\n\n");
      exit (EXIT_FAILURE);
   }
   for (i = 0; i < 256 && !feof (infp); i++) {
      inchar = fgetc (infp);
      if (!feof (infp)) rx01 [0153400 + i] = inchar & 0xFF;
   }
   fclose (infp);

   /*
      Skew sectors for output.
   */
   for (sequence = 0; sequence < (76 * 26); sequence++) {
      j = 128 * sequence; /* start of source byte */
      track  = sequence / 26 + 1;
      sector = ( 6 * (track - 1) + 2 * sequence + (sequence % 26)/13 ) % 26
               + 1;
      addr = 128 * (26 * (track - 1) + (sector - 1)); /* Starting byte */

#ifdef DEBUG
      /*
         Print track and sector number sequence to verify proper
         sector interleaving within a track, and sector skewing
         between tracks.
      */
      fprintf (stderr, "%4d: (%2d, %2d)\n", sequence, track, sector);
#endif

      for (i = 0; i < 128; i++) {
         raw01 [addr + i] = rx01 [i + j];
      }
   }

   /*
      Done creating disk; now write it to stdout.
   */
   for (i = 0; i < (26 * 128); i++) putchar (0xE5); /* Track 0 */
   for (i = 0; i < DISK_BYTES; i++) putchar (raw01[i]);

   exit (EXIT_SUCCESS);
}
